#include "bwt.h"
#include "defs.h"
#include "error.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

struct Cell {
	uint32_t left, right, level;
};

static uint8_t *sir;
static uint32_t sir_size;
static uint32_t* suffix_array;

static struct Cell* stack;
static uint32_t stack_size;
static uint32_t stack_head;

/* TODO: too ugly coding, remove */
#define SYMB(x) (sir[suffix_array[x]+level])

inline uint8_t choose_pivot(uint32_t left, uint32_t right, uint32_t level) {
	uint32_t mid = (left+right)/2;

	if (SYMB(left) > SYMB(mid)) {
		uint32_t x = suffix_array[left];
		suffix_array[left] = suffix_array[mid];
		suffix_array[mid] = x;
	}
	if (SYMB(mid) > SYMB(right)) {
		uint32_t x = suffix_array[mid];
		suffix_array[mid] = suffix_array[right];
		suffix_array[right] = x;
	}
	if (SYMB(left) > SYMB(mid)) {
		uint32_t x = suffix_array[left];
		suffix_array[left] = suffix_array[mid];
		suffix_array[mid] = x;
	}

	return SYMB(mid);
}

void push_stack(uint32_t left, uint32_t right, uint32_t level) {
	if (stack_head == stack_size) {
		struct Cell* temp = malloc(stack_size*2*sizeof(struct Cell));
		assert(temp != NULL);
		memcpy(temp, stack, stack_size*sizeof(struct Cell));

		free(stack);
		stack = temp;

		stack_size *= 2;
		error_print(1, "stack_size = %u", stack_size);
	}

	stack[stack_head].left = left;
	stack[stack_head].right = right;
	stack[stack_head].level = level;
	++stack_head;
}

void sort() {
	uint8_t pivot;
	uint32_t l, r;   /* left and right used for partioning */
	uint32_t a, b;

	uint32_t left, right, level;

	--stack_head;
	left = stack[stack_head].left;
	right = stack[stack_head].right;
	level = stack[stack_head].level;

	/* rejects invalid ranges */
	if (left >= sir_size || right >= sir_size)
		return;
	if (left >= right)
		return;
	if (left+1 == right) {
		/* if only two strings to compare use memcmp (much faster) */
		int value = memcmp(sir+suffix_array[left]+level,
				           sir+suffix_array[right]+level,
						   sir_size-level);
		if (value > 0) {
			uint32_t temp = suffix_array[right];
			suffix_array[right] = suffix_array[left];
			suffix_array[left] = temp;
		}

		return;
	}

	/* TODO: use a fallback algorithm when level is too big */
	if (level == sir_size)
		return;

	pivot = choose_pivot(left, right, level);

	l = left; r = right;
	while (l != r) {
		while (SYMB(l) <= pivot && l != r)
			++l;
		while (SYMB(r) > pivot && l != r)
			--r;

		{
			const uint32_t x = suffix_array[l];
			suffix_array[l] = suffix_array[r];
			suffix_array[r] = x;
		}
	}

	if (SYMB(r) > pivot && r != left)
		--r;
	a = r+1;    /* start of suffixes with first character > pivot */

	/* partitions suffixes <=1 pivot into 2 (< and ==) */
	/* avoid bug: r is always >= 0 */
	l = r;
	while (right >= r && r >= left) {
		if (SYMB(r) == pivot) {
			const uint32_t x = suffix_array[l];
			suffix_array[l] = suffix_array[r];
			suffix_array[r] = x;
			--l;
		}
		--r;
	}
	b = l;    /* end of suffixes with first character < pivot */

	push_stack(a, right, level);
	push_stack(left, b, level);
	push_stack(b+1, a-1, level+1);
}

/*
 * Burrows Wheeler Transformation -> forward
 * aloca memorie pentru output
 */
void bwt_forward(const uint8_t *in, uint32_t in_size, uint8_t **out,
                 uint32_t *out_size, uint32_t *index)
{
	uint8_t* output;
	uint32_t i;

	/* makes a copy of in string */
	sir = malloc(2*in_size);
	memcpy(sir, in, in_size);
	memcpy(sir+in_size, in, in_size);
	sir_size = in_size;

	suffix_array = malloc(in_size*sizeof(uint32_t));
	assert(suffix_array != NULL);
	for (i = 0; i < in_size; ++i)
		suffix_array[i] = i;		

	stack_size = 16;
	stack = malloc(stack_size*sizeof(struct Cell));
	stack_head = 0;

	push_stack(0, in_size-1, 0);
	while (stack_head > 0) {
		sort();
	}

	free(stack);

	// lets test sorting phase
	/*
	for (i = 0; i < in_size-1; ++i) {
		int value = memcmp(sir+suffix_array[i],
				           sir+suffix_array[i+1],
						   in_size);
		int j = 0;

		for (; j < in_size; ++j)
			printf("%c", sir[suffix_array[i]+j]);
		printf(" %d\n", value);
	}
	*/

	// frees memory from sir because it's no longer needed
	free(sir);

    output = (uint8_t *)malloc(in_size);
    assert(output != NULL);
    *out = output; *out_size = in_size;

    for(i = 0; i < in_size; ++ i)
        if(suffix_array[i] == 0)
            output[*index = i] = in[in_size-1];
        else
            output[i] = in[suffix_array[i]-1];

	free(suffix_array);
}

/*
 * Burrows Wheeler Transformation -> reverse
 * aloca memorie pentru output
 */
void bwt_reverse(const uint8_t *in, uint32_t in_size, uint8_t **out,
                 uint32_t *out_size, uint32_t index)
{
    uint32_t freq[256];
    uint32_t i, temp;
    uint32_t *where;
    uint8_t *output;

    for(i = 0; i < 256; ++ i)
        freq[i] = 0;
    for(i = 0; i < in_size; ++ i)
        freq[in[i]] ++;

    temp = 0;
    for(i = 0; i < 256; ++ i) {
        const uint32_t save = freq[i];
        freq[i] = temp, temp += save;
    }

    assert(temp == in_size);

    where = (uint32_t *)malloc(sizeof(uint32_t) * in_size);
    assert(where != NULL);

    for(i = 0; i < in_size; ++ i)
        where[freq[in[i]] ++] = i;

    output = (uint8_t *)malloc(in_size);
    assert(output != NULL);
    *out = output; *out_size = in_size;

    temp = index;
    for(i = 0; i < in_size; ++ i) {
        temp = where[temp];
        output[i] = in[temp];
    }
    assert(temp == index);

    free(where);
}
